@@ -21,13 +21,16 @@
 #include "mongoc-client-session-private.h"
 #include "mongoc-counters-private.h"
 #include "mongoc-error.h"
+#include "mongoc-error-private.h"
 #include "mongoc-log.h"
 #include "mongoc-trace-private.h"
 #include "mongoc-read-concern-private.h"
 #include "mongoc-util-private.h"
 #include "mongoc-write-concern-private.h"
 #include "mongoc-read-prefs-private.h"
+#include "mongoc-aggregate-private.h"
 
+#include <bson-dsl.h>
 
 #undef MONGOC_LOG_DOMAIN
 #define MONGOC_LOG_DOMAIN "cursor"
@@ -137,7 +140,13 @@ _mongoc_n_return (mongoc_cursor_t *cursor)
    /* if a specified limit exists, account for documents already returned. */
    if (limit > 0 && cursor->count) {
       int64_t remaining = limit - cursor->count;
-      BSON_ASSERT (remaining > 0);
+      /* remaining can be 0 if we have retrieved "limit" documents, but still
+       * have a cursor id: SERVER-21086. use nonzero batchSize to fetch final
+       * empty batch and trigger server to close cursor. */
+      if (remaining <= 0) {
+         return 1;
+      }
+
       n_return = BSON_MIN (n_return, remaining);
    }
 
@@ -157,8 +166,9 @@ _mongoc_set_cursor_ns (mongoc_cursor_t *cursor, const char *ns, uint32_t nslen)
 {
    const char *dot;
 
-   bson_strncpy (cursor->ns, ns, sizeof cursor->ns);
-   cursor->nslen = BSON_MIN (nslen, sizeof cursor->ns);
+   bson_free (cursor->ns);
+   cursor->ns = bson_strndup (ns, nslen);
+   cursor->nslen = nslen;
    dot = strstr (cursor->ns, ".");
 
    if (dot) {
@@ -230,17 +240,20 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
    mongoc_cursor_t *cursor;
    mongoc_topology_description_type_t td_type;
    uint32_t server_id;
+   mongoc_read_concern_t *read_concern_local = NULL;
    bson_error_t validate_err;
    const char *dollar_field;
    bson_iter_t iter;
 
    ENTRY;
 
-   BSON_ASSERT (client);
+   BSON_ASSERT_PARAM (client);
 
-   cursor = (mongoc_cursor_t *) bson_malloc0 (sizeof *cursor);
+   cursor = BSON_ALIGNED_ALLOC0 (mongoc_cursor_t);
    cursor->client = client;
    cursor->state = UNPRIMED;
+   cursor->client_generation = client->generation;
+   cursor->is_aggr_with_write_stage = false;
 
    bson_init (&cursor->opts);
    bson_init (&cursor->error_doc);
@@ -275,6 +288,18 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
          cursor->explicit_session = true;
       }
 
+      if (bson_iter_init_find (&iter, opts, "readConcern")) {
+         read_concern_local =
+            _mongoc_read_concern_new_from_iter (&iter, &cursor->error);
+
+         if (!read_concern_local) {
+            /* invalid read concern */
+            GOTO (finish);
+         }
+
+         read_concern = read_concern_local;
+      }
+
       /* true if there's a valid serverId or no serverId, false on err */
       if (!_mongoc_get_server_id_from_opts (opts,
                                             MONGOC_ERROR_CURSOR,
@@ -288,8 +313,13 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
          (void) mongoc_cursor_set_hint (cursor, server_id);
       }
 
-      bson_copy_to_excluding_noinit (
-         opts, &cursor->opts, "serverId", "sessionId", NULL);
+      // Selectively copy the options:
+      bsonBuildAppend (
+         cursor->opts,
+         insert (*opts,
+                 not(key ("serverId", "sessionId"),
+                     // Drop bypassDocumentValidation if it isn't true:
+                     allOf (key ("bypassDocumentValidation"), isFalse))));
    }
 
    if (_mongoc_client_session_in_txn (cursor->client_session)) {
@@ -301,15 +331,6 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
          GOTO (finish);
       }
 
-      if (user_prefs) {
-         bson_set_error (
-            &cursor->error,
-            MONGOC_ERROR_CURSOR,
-            MONGOC_ERROR_CURSOR_INVALID_CURSOR,
-            "Cannot set read preferences after starting transaction");
-         GOTO (finish);
-      }
-
       cursor->read_prefs =
          mongoc_read_prefs_copy (cursor->client_session->txn.opts.read_prefs);
 
@@ -359,6 +380,7 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
    (void) _mongoc_read_prefs_validate (cursor->read_prefs, &cursor->error);
 
 finish:
+   mongoc_read_concern_destroy (read_concern_local);
    mongoc_counter_cursors_active_inc ();
 
    RETURN (cursor);
@@ -418,19 +440,19 @@ _translate_query_opt (const char *query_field, const char **cmd_field, int *len)
 
 
 /* set up a new opt bson from older ways of specifying options.
- * slave_ok may be NULL.
+ * secondary_ok may be NULL.
  * error may be NULL.
  */
 void
 _mongoc_cursor_flags_to_opts (mongoc_query_flags_t qflags,
                               bson_t *opts, /* IN/OUT */
-                              bool *slave_ok /* OUT */)
+                              bool *secondary_ok /* OUT */)
 {
    ENTRY;
    BSON_ASSERT (opts);
 
-   if (slave_ok) {
-      *slave_ok = !!(qflags & MONGOC_QUERY_SLAVE_OK);
+   if (secondary_ok) {
+      *secondary_ok = !!(qflags & MONGOC_QUERY_SECONDARY_OK);
    }
 
    if (qflags & MONGOC_QUERY_TAILABLE_CURSOR) {
@@ -472,9 +494,9 @@ _mongoc_cursor_flags_to_opts (mongoc_query_flags_t qflags,
 
 /* Checks if the passed query was wrapped in a $query, and if so, parses the
  * query modifiers:
- * https://docs.mongodb.com/manual/reference/operator/query-modifier/
+ * https://www.mongodb.com/docs/manual/reference/operator/query-modifier/
  * and translates them to find command options:
- * https://docs.mongodb.com/manual/reference/command/find/
+ * https://www.mongodb.com/docs/manual/reference/command/find/
  * opts must be initialized, and may already have options set.
  * unwrapped must be uninitialized, and will be initialized at return.
  * Returns true if query was unwrapped. */
@@ -563,7 +585,7 @@ done:
 void
 mongoc_cursor_destroy (mongoc_cursor_t *cursor)
 {
-   char db[MONGOC_NAMESPACE_MAX];
+   char *db;
    ENTRY;
 
    if (!cursor) {
@@ -574,23 +596,30 @@ mongoc_cursor_destroy (mongoc_cursor_t *cursor)
       cursor->impl.destroy (&cursor->impl);
    }
 
+   /* Always close the socket for an exhaust cursor, even if the client was
+    * reset with mongoc_client_reset. That prevents further use of that socket.
+    */
    if (cursor->in_exhaust) {
       cursor->client->in_exhaust = false;
       if (cursor->state != DONE) {
-         /* The only way to stop an exhaust cursor is to kill the connection */
-         mongoc_cluster_disconnect_node (
-            &cursor->client->cluster, cursor->server_id, false, NULL);
+         /* The only way to stop an exhaust cursor is to kill the connection
+          */
+         mongoc_cluster_disconnect_node (&cursor->client->cluster,
+                                         cursor->server_id);
       }
-   } else if (cursor->cursor_id) {
-      bson_strncpy (db, cursor->ns, cursor->dblen + 1);
+   } else if (cursor->client_generation == cursor->client->generation) {
+      if (cursor->cursor_id) {
+         db = bson_strndup (cursor->ns, cursor->dblen);
 
-      _mongoc_client_kill_cursor (cursor->client,
-                                  cursor->server_id,
-                                  cursor->cursor_id,
-                                  cursor->operation_id,
-                                  db,
-                                  cursor->ns + cursor->dblen + 1,
-                                  cursor->client_session);
+         _mongoc_client_kill_cursor (cursor->client,
+                                     cursor->server_id,
+                                     cursor->cursor_id,
+                                     cursor->operation_id,
+                                     db,
+                                     cursor->ns + cursor->dblen + 1,
+                                     cursor->client_session);
+         bson_free (db);
+      }
    }
 
    if (cursor->client_session && !cursor->explicit_session) {
@@ -603,6 +632,7 @@ mongoc_cursor_destroy (mongoc_cursor_t *cursor)
 
    bson_destroy (&cursor->opts);
    bson_destroy (&cursor->error_doc);
+   bson_free (cursor->ns);
    bson_free (cursor);
 
    mongoc_counter_cursors_active_dec ();
@@ -621,6 +651,8 @@ _mongoc_cursor_fetch_stream (mongoc_cursor_t *cursor)
    ENTRY;
 
    if (cursor->server_id) {
+      /* We already did server selection once before. Reuse the prior
+       * selection to create a new stream on the same server. */
       server_stream =
          mongoc_cluster_stream_for_server (&cursor->client->cluster,
                                            cursor->server_id,
@@ -628,15 +660,32 @@ _mongoc_cursor_fetch_stream (mongoc_cursor_t *cursor)
                                            cursor->client_session,
                                            &reply,
                                            &cursor->error);
+      if (server_stream) {
+         /* Also restore whether primary read preference was forced by server
+          * selection */
+         server_stream->must_use_primary = cursor->must_use_primary;
+      }
    } else {
-      server_stream = mongoc_cluster_stream_for_reads (&cursor->client->cluster,
-                                                       cursor->read_prefs,
-                                                       cursor->client_session,
-                                                       &reply,
-                                                       &cursor->error);
+      server_stream =
+         cursor->is_aggr_with_write_stage
+            ? mongoc_cluster_stream_for_aggr_with_write (
+                 &cursor->client->cluster,
+                 cursor->read_prefs,
+                 cursor->client_session,
+                 &reply,
+                 &cursor->error)
+            : mongoc_cluster_stream_for_reads (&cursor->client->cluster,
+                                               cursor->read_prefs,
+                                               cursor->client_session,
+                                               &reply,
+                                               &cursor->error);
 
       if (server_stream) {
+         /* Remember the selected server_id and whether primary read mode was
+          * forced so that we can re-create an equivalent server_stream at a
+          * later time */
          cursor->server_id = server_stream->sd->id;
+         cursor->must_use_primary = server_stream->must_use_primary;
       }
    }
 
@@ -658,7 +707,7 @@ _mongoc_cursor_monitor_command (mongoc_cursor_t *cursor,
 {
    mongoc_client_t *client;
    mongoc_apm_command_started_t event;
-   char db[MONGOC_NAMESPACE_MAX];
+   char *db;
 
    ENTRY;
 
@@ -668,7 +717,7 @@ _mongoc_cursor_monitor_command (mongoc_cursor_t *cursor,
       RETURN (true);
    }
 
-   bson_strncpy (db, cursor->ns, cursor->dblen + 1);
+   db = bson_strndup (cursor->ns, cursor->dblen);
 
    mongoc_apm_command_started_init (&event,
                                     cmd,
@@ -678,10 +727,14 @@ _mongoc_cursor_monitor_command (mongoc_cursor_t *cursor,
                                     cursor->operation_id,
                                     &server_stream->sd->host,
                                     server_stream->sd->id,
+                                    &server_stream->sd->service_id,
+                                    server_stream->sd->server_connection_id,
+                                    NULL,
                                     client->apm_context);
 
    client->apm_callbacks.started (&event);
    mongoc_apm_command_started_cleanup (&event);
+   bson_free (db);
 
    RETURN (true);
 }
@@ -700,6 +753,8 @@ _mongoc_cursor_append_docs_array (mongoc_cursor_t *cursor,
    size_t keylen;
    const bson_t *doc;
 
+   BSON_UNUSED (cursor);
+
    while ((doc = bson_reader_read (response->reader, &eof))) {
       keylen = bson_uint32_to_string (i, &key, str, sizeof str);
       bson_append_document (docs, key, (int) keylen, doc);
@@ -720,8 +775,6 @@ _mongoc_cursor_monitor_succeeded (mongoc_cursor_t *cursor,
    bson_t docs_array;
    mongoc_apm_command_succeeded_t event;
    mongoc_client_t *client;
-   bson_t reply;
-   bson_t reply_cursor;
 
    ENTRY;
 
@@ -737,16 +790,14 @@ _mongoc_cursor_monitor_succeeded (mongoc_cursor_t *cursor,
    bson_init (&docs_array);
    _mongoc_cursor_append_docs_array (cursor, &docs_array, response);
 
-   bson_init (&reply);
-   bson_append_int32 (&reply, "ok", 2, 1);
-   bson_append_document_begin (&reply, "cursor", 6, &reply_cursor);
-   bson_append_int64 (&reply_cursor, "id", 2, mongoc_cursor_get_id (cursor));
-   bson_append_utf8 (&reply_cursor, "ns", 2, cursor->ns, cursor->nslen);
-   bson_append_array (&reply_cursor,
-                      first_batch ? "firstBatch" : "nextBatch",
-                      first_batch ? 10 : 9,
-                      &docs_array);
-   bson_append_document_end (&reply, &reply_cursor);
+   bsonBuildDecl (reply,
+                  kv ("ok", int32 (1)),
+                  kv ("cursor",
+                      doc (kv ("id", int64 (mongoc_cursor_get_id (cursor))),
+                           kv ("ns", utf8_w_len (cursor->ns, cursor->nslen)),
+                           kv (first_batch ? "firstBatch" : "nextBatch",
+                               bsonArray (docs_array)))));
+
    bson_destroy (&docs_array);
 
    mongoc_apm_command_succeeded_init (&event,
@@ -757,6 +808,9 @@ _mongoc_cursor_monitor_succeeded (mongoc_cursor_t *cursor,
                                       cursor->operation_id,
                                       &stream->sd->host,
                                       stream->sd->id,
+                                      &stream->sd->service_id,
+                                      stream->sd->server_connection_id,
+                                      false,
                                       client->apm_context);
 
    client->apm_callbacks.succeeded (&event);
@@ -776,7 +830,6 @@ _mongoc_cursor_monitor_failed (mongoc_cursor_t *cursor,
 {
    mongoc_apm_command_failed_t event;
    mongoc_client_t *client;
-   bson_t reply;
 
    ENTRY;
 
@@ -789,8 +842,7 @@ _mongoc_cursor_monitor_failed (mongoc_cursor_t *cursor,
    /* we sent OP_QUERY/OP_GETMORE, fake a reply to find/getMore command:
     * {ok: 0}
     */
-   bson_init (&reply);
-   bson_append_int32 (&reply, "ok", 2, 0);
+   bsonBuildDecl (reply, kv ("ok", int32 (0)));
 
    mongoc_apm_command_failed_init (&event,
                                    duration,
@@ -801,6 +853,9 @@ _mongoc_cursor_monitor_failed (mongoc_cursor_t *cursor,
                                    cursor->operation_id,
                                    &stream->sd->host,
                                    stream->sd->id,
+                                   &stream->sd->service_id,
+                                   stream->sd->server_connection_id,
+                                   false,
                                    client->apm_context);
 
    client->apm_callbacks.failed (&event);
@@ -830,12 +885,12 @@ _mongoc_cursor_monitor_failed (mongoc_cursor_t *cursor,
 bool
 _mongoc_cursor_opts_to_flags (mongoc_cursor_t *cursor,
                               mongoc_server_stream_t *stream,
-                              mongoc_query_flags_t *flags /* OUT */)
+                              int32_t *flags /* OUT */)
 {
    bson_iter_t iter;
    const char *key;
 
-   *flags = MONGOC_QUERY_NONE;
+   *flags = MONGOC_OP_QUERY_FLAG_NONE;
 
    if (!bson_iter_init (&iter, &cursor->opts)) {
       bson_set_error (&cursor->error,
@@ -849,27 +904,27 @@ _mongoc_cursor_opts_to_flags (mongoc_cursor_t *cursor,
       key = bson_iter_key (&iter);
 
       if (!strcmp (key, MONGOC_CURSOR_ALLOW_PARTIAL_RESULTS)) {
-         ADD_FLAG (flags, MONGOC_QUERY_PARTIAL);
+         ADD_FLAG (flags, MONGOC_OP_QUERY_FLAG_PARTIAL);
       } else if (!strcmp (key, MONGOC_CURSOR_AWAIT_DATA)) {
-         ADD_FLAG (flags, MONGOC_QUERY_AWAIT_DATA);
+         ADD_FLAG (flags, MONGOC_OP_QUERY_FLAG_AWAIT_DATA);
       } else if (!strcmp (key, MONGOC_CURSOR_EXHAUST)) {
-         ADD_FLAG (flags, MONGOC_QUERY_EXHAUST);
+         ADD_FLAG (flags, MONGOC_OP_QUERY_FLAG_EXHAUST);
       } else if (!strcmp (key, MONGOC_CURSOR_NO_CURSOR_TIMEOUT)) {
-         ADD_FLAG (flags, MONGOC_QUERY_NO_CURSOR_TIMEOUT);
+         ADD_FLAG (flags, MONGOC_OP_QUERY_FLAG_NO_CURSOR_TIMEOUT);
       } else if (!strcmp (key, MONGOC_CURSOR_OPLOG_REPLAY)) {
-         ADD_FLAG (flags, MONGOC_QUERY_OPLOG_REPLAY);
+         ADD_FLAG (flags, MONGOC_OP_QUERY_FLAG_OPLOG_REPLAY);
       } else if (!strcmp (key, MONGOC_CURSOR_TAILABLE)) {
-         ADD_FLAG (flags, MONGOC_QUERY_TAILABLE_CURSOR);
+         ADD_FLAG (flags, MONGOC_OP_QUERY_FLAG_TAILABLE_CURSOR);
       }
    }
 
-   if (cursor->slave_ok) {
-      *flags |= MONGOC_QUERY_SLAVE_OK;
+   if (cursor->secondary_ok) {
+      *flags |= MONGOC_OP_QUERY_FLAG_SECONDARY_OK;
    } else if (cursor->server_id &&
               (stream->topology_type == MONGOC_TOPOLOGY_RS_WITH_PRIMARY ||
                stream->topology_type == MONGOC_TOPOLOGY_RS_NO_PRIMARY) &&
               stream->sd->type != MONGOC_SERVER_RS_PRIMARY) {
-      *flags |= MONGOC_QUERY_SLAVE_OK;
+      *flags |= MONGOC_OP_QUERY_FLAG_SECONDARY_OK;
    }
 
    return true;
@@ -879,22 +934,22 @@ bool
 _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
                             const bson_t *command,
                             const bson_t *opts,
-                            bson_t *reply)
+                            bson_t *reply,
+                            bool retry_prohibited)
 {
-   mongoc_cluster_t *cluster;
    mongoc_server_stream_t *server_stream;
    bson_iter_t iter;
    mongoc_cmd_parts_t parts;
    const char *cmd_name;
    bool is_primary;
    mongoc_read_prefs_t *prefs = NULL;
-   char db[MONGOC_NAMESPACE_MAX];
+   char *db = NULL;
    mongoc_session_opt_t *session_opts;
    bool ret = false;
+   bool is_retryable = true;
 
    ENTRY;
 
-   cluster = &cursor->client->cluster;
    mongoc_cmd_parts_init (
       &parts, cursor->client, db, MONGOC_QUERY_NONE, command);
    parts.is_read_command = true;
@@ -916,13 +971,15 @@ _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
                          "Invalid BSON in opts document");
          GOTO (done);
       }
-      if (!mongoc_cmd_parts_append_opts (&parts,
-                                         &iter,
-                                         server_stream->sd->max_wire_version,
-                                         &cursor->error)) {
+      if (!mongoc_cmd_parts_append_opts (&parts, &iter, &cursor->error)) {
          _mongoc_bson_init_if_set (reply);
          GOTO (done);
       }
+      if (_mongoc_cursor_get_opt_bool (cursor, MONGOC_CURSOR_EXHAUST)) {
+         MONGOC_WARNING (
+            "exhaust cursors not supported with OP_MSG, using normal "
+            "cursor instead");
+      }
    }
 
    if (parts.assembled.session) {
@@ -948,44 +1005,70 @@ _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
       mongoc_session_opts_destroy (session_opts);
    }
 
-   if (!mongoc_cmd_parts_set_read_concern (&parts,
-                                           cursor->read_concern,
-                                           server_stream->sd->max_wire_version,
-                                           &cursor->error)) {
+   if (!mongoc_cmd_parts_set_read_concern (
+          &parts, cursor->read_concern, &cursor->error)) {
       _mongoc_bson_init_if_set (reply);
       GOTO (done);
    }
 
-   bson_strncpy (db, cursor->ns, cursor->dblen + 1);
+   db = bson_strndup (cursor->ns, cursor->dblen);
    parts.assembled.db_name = db;
 
-   if (!_mongoc_cursor_opts_to_flags (
-          cursor, server_stream, &parts.user_query_flags)) {
-      _mongoc_bson_init_if_set (reply);
-      GOTO (done);
+   {
+      int32_t flags;
+      if (!_mongoc_cursor_opts_to_flags (cursor, server_stream, &flags)) {
+         _mongoc_bson_init_if_set (reply);
+         GOTO (done);
+      }
+      parts.user_query_flags = (mongoc_query_flags_t) flags;
+   }
+
+   /* Exhaust cursors with OP_MSG not yet supported; fallback to normal cursor.
+    * user_query_flags is unused in OP_MSG, so this technically has no effect,
+    * but is done anyways to ensure the query flags match handling of options.
+    */
+   if (parts.user_query_flags & MONGOC_QUERY_EXHAUST) {
+      parts.user_query_flags ^= MONGOC_QUERY_EXHAUST;
    }
 
    /* we might use mongoc_cursor_set_hint to target a secondary but have no
     * read preference, so the secondary rejects the read. same if we have a
     * direct connection to a secondary (topology type "single"). with
-    * OP_QUERY we handle this by setting slaveOk. here we use $readPreference.
+    * OP_QUERY we handle this by setting secondaryOk. here we use
+    * $readPreference.
     */
    cmd_name = _mongoc_get_command_name (command);
    is_primary =
       !cursor->read_prefs || cursor->read_prefs->mode == MONGOC_READ_PRIMARY;
 
-   if (strcmp (cmd_name, "getMore") != 0 &&
-       server_stream->sd->max_wire_version >= WIRE_VERSION_OP_MSG &&
-       is_primary && parts.user_query_flags & MONGOC_QUERY_SLAVE_OK) {
+   if (strcmp (cmd_name, "getMore") != 0 && is_primary &&
+       parts.user_query_flags & MONGOC_QUERY_SECONDARY_OK) {
       parts.read_prefs = prefs =
          mongoc_read_prefs_new (MONGOC_READ_PRIMARY_PREFERRED);
    } else {
       parts.read_prefs = cursor->read_prefs;
    }
 
+   is_retryable = _is_retryable_read (&parts, server_stream);
+   if (!strcmp (cmd_name, "getMore")) {
+      is_retryable = false;
+   }
+   if (!strcmp (cmd_name, "aggregate")) {
+      bson_iter_t pipeline_iter;
+      if (bson_iter_init_find (&pipeline_iter, command, "pipeline") &&
+          BSON_ITER_HOLDS_ARRAY (&pipeline_iter) &&
+          bson_iter_recurse (&pipeline_iter, &pipeline_iter)) {
+         if (_has_write_key (&pipeline_iter)) {
+            is_retryable = false;
+         }
+      }
+   }
+   if (is_retryable && retry_prohibited) {
+      is_retryable = false;
+   }
+
    if (cursor->write_concern &&
-       !mongoc_write_concern_is_default (cursor->write_concern) &&
-       server_stream->sd->max_wire_version >= WIRE_VERSION_CMD_WRITE_CONCERN) {
+       !mongoc_write_concern_is_default (cursor->write_concern)) {
       parts.assembled.is_acknowledged =
          mongoc_write_concern_is_acknowledged (cursor->write_concern);
       mongoc_write_concern_append (cursor->write_concern, &parts.extra);
@@ -996,8 +1079,37 @@ _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
       GOTO (done);
    }
 
+retry:
    ret = mongoc_cluster_run_command_monitored (
-      cluster, &parts.assembled, reply, &cursor->error);
+      &cursor->client->cluster, &parts.assembled, reply, &cursor->error);
+
+   if (ret) {
+      memset (&cursor->error, 0, sizeof (bson_error_t));
+   }
+
+   if (is_retryable &&
+       _mongoc_read_error_get_type (ret, &cursor->error, reply) ==
+          MONGOC_READ_ERR_RETRY) {
+      is_retryable = false;
+
+      mongoc_server_stream_cleanup (server_stream);
+
+      BSON_ASSERT (!cursor->is_aggr_with_write_stage &&
+                   "Cannot attempt a retry on an aggregate operation that "
+                   "contains write stages");
+      server_stream = mongoc_cluster_stream_for_reads (&cursor->client->cluster,
+                                                       cursor->read_prefs,
+                                                       cursor->client_session,
+                                                       reply,
+                                                       &cursor->error);
+
+      if (server_stream) {
+         cursor->server_id = server_stream->sd->id;
+         parts.assembled.server_stream = server_stream;
+         bson_destroy (reply);
+         GOTO (retry);
+      }
+   }
 
    if (cursor->error.domain) {
       bson_destroy (&cursor->error_doc);
@@ -1020,6 +1132,7 @@ done:
    mongoc_server_stream_cleanup (server_stream);
    mongoc_cmd_parts_cleanup (&parts);
    mongoc_read_prefs_destroy (prefs);
+   bson_free (db);
 
    return ret;
 }
@@ -1142,6 +1255,14 @@ mongoc_cursor_next (mongoc_cursor_t *cursor, const bson_t **bson)
 
    TRACE ("cursor_id(%" PRId64 ")", cursor->cursor_id);
 
+   if (cursor->client_generation != cursor->client->generation) {
+      bson_set_error (&cursor->error,
+                      MONGOC_ERROR_CURSOR,
+                      MONGOC_ERROR_CURSOR_INVALID_CURSOR,
+                      "Cannot advance cursor after client reset");
+      RETURN (false);
+   }
+
    if (bson) {
       *bson = NULL;
    }
@@ -1227,7 +1348,8 @@ mongoc_cursor_more (mongoc_cursor_t *cursor)
 void
 mongoc_cursor_get_host (mongoc_cursor_t *cursor, mongoc_host_list_t *host)
 {
-   mongoc_server_description_t *description;
+   mongoc_server_description_t const *description;
+   mc_shared_tpld td;
 
    BSON_ASSERT (cursor);
    BSON_ASSERT (host);
@@ -1239,16 +1361,16 @@ mongoc_cursor_get_host (mongoc_cursor_t *cursor, mongoc_host_list_t *host)
       return;
    }
 
-   description = mongoc_topology_server_by_id (
-      cursor->client->topology, cursor->server_id, &cursor->error);
+   td = mc_tpld_take_ref (cursor->client->topology);
+   description = mongoc_topology_description_server_by_id_const (
+      td.ptr, cursor->server_id, &cursor->error);
+   mc_tpld_drop_ref (&td);
    if (!description) {
       return;
    }
 
    *host = description->host;
 
-   mongoc_server_description_destroy (description);
-
    EXIT;
 }
 
@@ -1260,7 +1382,7 @@ mongoc_cursor_clone (const mongoc_cursor_t *cursor)
 
    BSON_ASSERT (cursor);
 
-   _clone = (mongoc_cursor_t *) bson_malloc0 (sizeof *_clone);
+   _clone = BSON_ALIGNED_ALLOC0 (mongoc_cursor_t);
 
    _clone->client = cursor->client;
    _clone->nslen = cursor->nslen;
@@ -1286,7 +1408,7 @@ mongoc_cursor_clone (const mongoc_cursor_t *cursor)
    bson_copy_to (&cursor->opts, &_clone->opts);
    bson_init (&_clone->error_doc);
 
-   bson_strncpy (_clone->ns, cursor->ns, sizeof _clone->ns);
+   _clone->ns = bson_strdup (cursor->ns);
 
    /* copy the context functions by default. */
    memcpy (&_clone->impl, &cursor->impl, sizeof (cursor->impl));
@@ -1330,9 +1452,32 @@ void
 mongoc_cursor_set_batch_size (mongoc_cursor_t *cursor, uint32_t batch_size)
 {
    BSON_ASSERT (cursor);
-
-   _mongoc_cursor_set_opt_int64 (
-      cursor, MONGOC_CURSOR_BATCH_SIZE, (int64_t) batch_size);
+   bson_iter_t iter;
+   if (!bson_iter_init_find (&iter, &cursor->opts, MONGOC_CURSOR_BATCH_SIZE)) {
+      bson_append_int64 (&cursor->opts,
+                         MONGOC_CURSOR_BATCH_SIZE,
+                         MONGOC_CURSOR_BATCH_SIZE_LEN,
+                         batch_size);
+   } else if (BSON_ITER_HOLDS_INT64 (&iter)) {
+      bson_iter_overwrite_int64 (&iter, (int64_t) batch_size);
+   } else if (BSON_ITER_HOLDS_INT32 (&iter)) {
+      if (!bson_in_range_int32_t_unsigned (batch_size)) {
+         MONGOC_WARNING ("unable to overwrite stored int32 batchSize with "
+                         "out-of-range value %" PRIu32,
+                         batch_size);
+         return;
+      }
+      bson_iter_overwrite_int32 (&iter, (int32_t) batch_size);
+   } else if (BSON_ITER_HOLDS_DOUBLE (&iter)) {
+      bson_iter_overwrite_double (&iter, (double) batch_size);
+   } else if (BSON_ITER_HOLDS_DECIMAL128 (&iter)) {
+      bson_decimal128_t val;
+      val.high = 0x3040000000000000;
+      val.low = (uint64_t) batch_size;
+      bson_iter_overwrite_decimal128 (&iter, &val);
+   } else {
+      MONGOC_WARNING ("unable to overwrite non-numeric stored batchSize");
+   }
 }
 
 
@@ -1465,17 +1610,15 @@ mongoc_cursor_new_from_command_reply (mongoc_client_t *client,
    bson_t cmd = BSON_INITIALIZER;
    bson_t opts = BSON_INITIALIZER;
 
-   BSON_ASSERT (client);
+   BSON_ASSERT_PARAM (client);
    BSON_ASSERT (reply);
    /* options are passed through by adding them to reply. */
-   bson_copy_to_excluding_noinit (reply,
-                                  &opts,
-                                  "cursor",
-                                  "ok",
-                                  "operationTime",
-                                  "$clusterTime",
-                                  "$gleStats",
-                                  NULL);
+   bsonBuildAppend (
+      *reply,
+      insert (
+         opts,
+         not(key (
+            "cursor", "ok", "operationTime", "$clusterTime", "$gleStats"))));
 
    if (server_id) {
       bson_append_int64 (&opts, "serverId", 8, server_id);
@@ -1497,7 +1640,7 @@ mongoc_cursor_new_from_command_reply_with_opts (mongoc_client_t *client,
    mongoc_cursor_t *cursor;
    bson_t cmd = BSON_INITIALIZER;
 
-   BSON_ASSERT (client);
+   BSON_ASSERT_PARAM (client);
    BSON_ASSERT (reply);
 
    cursor = _mongoc_cursor_cmd_new_from_reply (client, &cmd, opts, reply);
@@ -1559,6 +1702,8 @@ _mongoc_cursor_response_read (mongoc_cursor_t *cursor,
 
    ENTRY;
 
+   BSON_UNUSED (cursor);
+
    if (bson_iter_next (&response->batch_iter) &&
        BSON_ITER_HOLDS_DOCUMENT (&response->batch_iter)) {
       bson_iter_document (&response->batch_iter, &data_len, &data);
@@ -1582,7 +1727,8 @@ _mongoc_cursor_response_refresh (mongoc_cursor_t *cursor,
 
    /* server replies to find / aggregate with {cursor: {id: N, firstBatch: []}},
     * to getMore command with {cursor: {id: N, nextBatch: []}}. */
-   if (_mongoc_cursor_run_command (cursor, command, opts, &response->reply) &&
+   if (_mongoc_cursor_run_command (
+          cursor, command, opts, &response->reply, false) &&
        _mongoc_cursor_start_reading_response (cursor, response)) {
       return;
    }
@@ -1595,7 +1741,6 @@ _mongoc_cursor_response_refresh (mongoc_cursor_t *cursor,
    }
 }
 
-
 void
 _mongoc_cursor_prepare_getmore_command (mongoc_cursor_t *cursor,
                                         bson_t *command)
@@ -1603,8 +1748,9 @@ _mongoc_cursor_prepare_getmore_command (mongoc_cursor_t *cursor,
    const char *collection;
    int collection_len;
    int64_t batch_size;
+   bson_iter_t iter;
    bool await_data;
-   int32_t max_await_time_ms;
+   int64_t max_await_time_ms;
 
    ENTRY;
 
@@ -1624,6 +1770,29 @@ _mongoc_cursor_prepare_getmore_command (mongoc_cursor_t *cursor,
                          abs (_mongoc_n_return (cursor)));
    }
 
+   if (bson_iter_init_find (&iter, &cursor->opts, MONGOC_CURSOR_COMMENT) &&
+       bson_iter_value (&iter)->value_type != BSON_TYPE_EOD) {
+      const bson_value_t *comment = bson_iter_value (&iter);
+      mongoc_server_stream_t *server_stream;
+
+      /* CRUD spec: If a comment is provided, drivers MUST attach this comment
+       * to all subsequent getMore commands run on the same cursor for server
+       * versions 4.4 and above. For server versions below 4.4 drivers MUST NOT
+       * attach a comment to getMore commands.
+       *
+       * Since this function has no error reporting, we also no-op if we cannot
+       * fetch a stream. */
+      server_stream = _mongoc_cursor_fetch_stream (cursor);
+
+      if (server_stream != NULL &&
+          server_stream->sd->max_wire_version >= WIRE_VERSION_4_4) {
+         bson_append_value (
+            command, MONGOC_CURSOR_COMMENT, MONGOC_CURSOR_COMMENT_LEN, comment);
+      }
+
+      mongoc_server_stream_cleanup (server_stream);
+   }
+
    /* Find, getMore And killCursors Commands Spec: "In the case of a tailable
       cursor with awaitData == true the driver MUST provide a Cursor level
       option named maxAwaitTimeMS (See CRUD specification for details). The
@@ -1634,12 +1803,11 @@ _mongoc_cursor_prepare_getmore_command (mongoc_cursor_t *cursor,
    await_data = _mongoc_cursor_get_opt_bool (cursor, MONGOC_CURSOR_TAILABLE) &&
                 _mongoc_cursor_get_opt_bool (cursor, MONGOC_CURSOR_AWAIT_DATA);
 
-
    if (await_data) {
-      max_await_time_ms =
-         (int32_t) mongoc_cursor_get_max_await_time_ms (cursor);
+      max_await_time_ms = _mongoc_cursor_get_opt_int64 (
+         cursor, MONGOC_CURSOR_MAX_AWAIT_TIME_MS, 0);
       if (max_await_time_ms) {
-         bson_append_int32 (command,
+         bson_append_int64 (command,
                             MONGOC_CURSOR_MAX_TIME_MS,
                             MONGOC_CURSOR_MAX_TIME_MS_LEN,
                             max_await_time_ms);
